@@ -4,7 +4,7 @@
 # Univention Management Console
 #  Univention Directory Manager Module
 #
-# Copyright 2019-2021 Univention GmbH
+# Copyright 2019-2022 Univention GmbH
 #
 # https://www.univention.de/
 #
@@ -305,6 +305,9 @@ class UDM(Client):
 		for module in self.modules(name):
 			return module
 
+	def get_object(self, object_type, dn):
+		return self.get(object_type).get(dn)
+
 	def __repr__(self):
 		return 'UDM(uri={}, username={}, password=****, version={})'.format(self.uri, self.username, self._api_version)
 
@@ -339,12 +342,14 @@ class Module(Client):
 		# TODO: use a link relation instead of a search
 		for obj in self.search(position=dn, scope='base'):
 			return obj.open()
+		raise NotFound(404, 'Wrong object type!?', None)  # FIXME: object exists but is of different module. should be fixed on the server.
 
 	def get_by_entry_uuid(self, uuid):
 		# TODO: use a link relation instead of a search
 		#return self.udm.get_by_uuid(uuid)
 		for obj in self.search(filter={'entryUUID': uuid}, scope='base'):
 			return obj.open()
+		raise NotFound(404, 'Wrong object type!?', None)  # FIXME: object exists but is of different module. should be fixed on the server.
 
 	def get_by_id(self, dn):
 		# TODO: Needed?
@@ -375,6 +380,34 @@ class Module(Client):
 				dn = objself['name']
 				yield ShallowObject(self.udm, dn, uri)
 
+	def get_layout(self):
+		self.load_relations()
+		return self.udm.client.resolve_relation(self.relations, 'udm:layout').get('layout')
+
+	def get_properties(self):
+		self.load_relations()
+		return self.udm.client.resolve_relation(self.relations, 'udm:properties').get('properties')
+
+	def get_property_choices(self, property):
+		self.load_relations()
+		relations = self.udm.client.resolve_relation(self.relations, 'udm:properties')
+		return self.udm.client.resolve_relation(relations, 'udm:property-choices', name=property).get('choices')
+
+	def policy_result(self, policy_module, position, policy=None):
+		self.load_relations()
+		policy_result = self.udm.client.resolve_relation(self.relations, 'udm:policy-result', name=policy_module, template={'position': position, 'policy': policy})
+		policy_result.pop('_links', None)
+		policy_result.pop('_embedded', None)
+		return policy_result
+
+	def get_report_types(self):
+		self.load_relations()
+		return [x['name'] for x in self.udm.client.get_relations(self.relations, 'udm:report', template={'dn': ''}) if x.get('name')]
+
+	def create_report(self, report_type, object_dns):
+		self.load_relations()
+		return self.udm.client.resolve_relation(self.relations, 'udm:report', name=report_type, template={'dn': object_dns})
+
 
 class ShallowObject(Client):
 
@@ -400,7 +433,7 @@ class References(object):
 	def __getitem__(self, item):
 		return [
 			ShallowObject(self.obj.udm, x['name'], x['href'])
-			for x in self.udm.get_relations(self.obj.hal, 'udm:object/property/reference/%s' % (item,))
+			for x in self.udm.client.get_relations(self.obj.hal, 'udm:object/property/reference/%s' % (item,))
 		]
 
 	def __getattribute__(self, key):
@@ -540,7 +573,7 @@ class Object(Client):
 	def _follow_redirection(self, response, reload=True):
 		location = None
 		# python-requests doesn't follow redirects for 201
-		if response.response.status_code == 201 and 'Location'in response.response.headers:
+		if response.response.status_code == 201 and 'Location' in response.response.headers:
 			location = response.response.headers['Location']
 			response = self.client.make_request('GET', location, allow_redirects=False)
 
@@ -564,3 +597,24 @@ class Object(Client):
 		self.hal = copy.deepcopy(obj.hal)
 		self.etag = obj.etag
 		self.last_modified = obj.last_modified
+
+	def generate_service_specific_password(self, service):
+		uri = self.client.get_relation(self.hal, 'udm:service-specific-password')['href']
+		response = self.client.make_request('POST', uri, data={"service": service})
+		return response.data.get('password', None)
+
+	def get_layout(self):
+		return self.udm.client.resolve_relation(self.hal, 'udm:layout').get('layout')
+
+	def get_properties(self):
+		return self.udm.client.resolve_relation(self.hal, 'udm:properties').get('properties')
+
+	def get_property_choices(self, property):
+		hal = self.udm.client.resolve_relation(self.hal, 'udm:properties')
+		return self.udm.client.resolve_relation(hal, 'udm:property-choices', name=property).get('choices')
+
+	def policy_result(self, policy_module, policy=None):
+		policy_result = self.udm.client.resolve_relation(self.hal, 'udm:policy-result', name=policy_module, template={'policy': policy})
+		policy_result.pop('_links', None)
+		policy_result.pop('_embedded', None)
+		return policy_result
